module KoL.Api where

import Prelude hiding (read, catch)

import KoL.Connection
import KoL.Util
import KoL.UtilTypes

import Text.JSON

import Control.Exception
import Control.Applicative

import qualified Data.ByteString.Char8

doAPI_DEBUG = return
-- doAPI_DEBUG x = putStrLn $ "API DEBUG: " ++ x

getCharStatusObj ref = do
	doAPI_DEBUG "getCharStatusObj"
	Ok jscomb <- decodeStrict <$> readstatus ref
	let Ok jsobj = valFromObj "status" jscomb
	return jsobj

getInventoryObj ref = do
	doAPI_DEBUG "getInventoryObj"
	Ok jscomb <- decodeStrict <$> readstatus ref
	let Ok jsobj = valFromObj "inventory" jscomb
	return jsobj

getInventoryCounts ref = do
	doAPI_DEBUG "getInventoryCounts"
	jsobj <- getInventoryObj ref
	let strcounts = fromJSObject jsobj
	let get_value name = i
		where i = case valFromObj name jsobj of
			Ok i -> i
			_ -> read_e j
				where Ok j = valFromObj name jsobj
	let counts = map (\(x, _y) -> (read_e x :: Integer, (get_value x) :: Integer)) strcounts
	return counts

data ApiInfo = ApiInfo {
	charName :: String,
	turnsplayed :: Integer,
	ascension :: Integer,
	daysthisrun :: Integer,
	pwd :: String
} -- deriving (Read, Show, Data, Typeable)

rawGetApiInfo ref jsontext = do
	let Ok jscomb = decodeStrict jsontext
	let Ok jsobj = valFromObj "status" jscomb
	let getstr what = case valFromObj what jsobj of
		Ok (JSString s) -> fromJSString s
		_ -> throw $ InternalError $ "Error parsing API text " ++ what
	let getnum what = case valFromObj what jsobj of
		Ok (JSString s) -> jss where
			Just jss = read_as (fromJSString s)
		Ok (JSRational _ r) -> round r
		Ok JSNull -> 0 -- HACK for when rollover is broken and not set yet
		_ -> throw $ InternalError $ "Error parsing API number " ++ what
	return ApiInfo { charName = getstr "name", turnsplayed = getnum "turnsplayed", ascension = getnum "ascensions" + 1, daysthisrun = getnum "daysthisrun", pwd = getstr "pwd" }

getApiInfo ref = do
	doAPI_DEBUG "getApiInfo"
	jsontext <- readstatus ref
	rawGetApiInfo ref jsontext

getStatus ref = do
	doAPI_DEBUG "getStatus"
	jsonstatusinv <- Data.ByteString.Char8.unpack <$> nochangeGetPageRawNoScripts ("/api.php?what=status,inventory&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json") ref
	case decodeStrict jsonstatusinv :: Result (JSObject JSValue) of
		Ok _ -> return jsonstatusinv
		_ -> do
			putStrLn $ "API status,inventory returned:\n  ===\n\n" ++ jsonstatusinv ++ "\n\n  ===\n\n"
			throwIO $ ApiPageException jsonstatusinv

getItemInfoObj itemid ref = do
	doAPI_DEBUG "getItemInfoObj"
	jsonobj <- Data.ByteString.Char8.unpack <$> nochangeGetPageRawNoScripts ("/api.php?what=item&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json&id=" ++ show itemid) ref
	case decodeStrict jsonobj :: Result JSValue of
		Ok x -> return x
		_ -> do
			putStrLn $ "API item returned:\n  ===\n\n" ++ jsonobj ++ "\n\n  ===\n\n"
			throwIO $ ApiPageException jsonobj

asyncGetItemInfoObj itemid ref = do
	f <- asyncNochangeGetPageRawNoScripts ("/api.php?what=item&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json&id=" ++ show itemid) ref
	return $ do
		jsonobj <- Data.ByteString.Char8.unpack <$> f
		case decodeStrict jsonobj :: Result JSValue of
			Ok x -> return x
			_ -> do
				putStrLn $ "API item returned:\n  ===\n\n" ++ jsonobj ++ "\n\n  ===\n\n"
				throwIO $ ApiPageException jsonobj
