module KoL.Http where

-- TODO: Merge with HttpLowlevel or at least restructure

import Prelude hiding (read, catch)

import Logging

import KoL.Util
import KoL.UtilTypes
import KoL.Api

import Control.Monad
import Control.Applicative
import Data.Maybe
import Data.Time
import Data.IORef
import KoL.HttpLowlevel
import Network.URI
import Control.Concurrent
import Control.Exception
import qualified Data.ByteString
import qualified Data.ByteString.Char8

doHTTP_DEBUG _ = return ()
-- doHTTP_DEBUG x = putStrLn $ "HTTP DEBUG: " ++ x

getHTTPFileData useragent url = do
	(_, body, _, _) <- doHTTPreq (mkreq useragent Nothing url Nothing True)
	return $ Data.ByteString.Char8.unpack body

postHTTPFileData useragent url params = do
	(_, body, _, _) <- doHTTPreq (mkreq useragent Nothing url (Just params) True)
	return $ Data.ByteString.Char8.unpack body

-- TODO: combine these three

internalKolHttpsRequest url params cu noredirect = do
	let (cookie, useragent, host, getconn) = cu
	let Just absuri = url `relativeTo` host
	(absuri, body, hdrs, code) <- doHTTPSreq (mkreq useragent cookie absuri params True)
	let addheaders = filter (\(x, _y) -> x == "Set-Cookie") hdrs
	return (body, absuri, addheaders)


internalKolRequest url params cu noredirect = do
	let (cookie, useragent, host, getconn) = cu
	let Just absuri = url `relativeTo` host
	putStrLn $ "DEBUG: single-req" ++ show absuri
	(absuri, body, hdrs, code) <- doHTTPreq (mkreq useragent cookie absuri params True)

	cookie <- case lookup "Set-Cookie" hdrs of
		Just hdrstr -> do
			let cookie = takeWhile (/= ';') hdrstr
			putStrLn $ "set-cookie_single: " ++ cookie
			return $ Just cookie
		Nothing -> return cookie
	if noredirect
		then return (body, absuri, hdrs)
		else case code of
			(3, _, _) -> do
				case lookup "Location" hdrs of
					Just x -> do
						let addheaders = filter (\(x, _y) -> x == "Set-Cookie") hdrs
						case parseURI x of
							Nothing -> do
								let remapper x
									| (x == ' ') = '+'
									| otherwise = x
								Just to <- case parseURIReference x of
									Just x -> return $ Just x
									_ -> do -- Workaround for broken KoL redirect, e.g. trade counter-offer doesn't urlencode()
										putStrLn $ "parse failure, invalid URL received from server: " ++ (show x) ++ " [Server bug, shout at Jick or CDMoyer!]"
										return $ parseURIReference (map remapper x)
-- 								let Just to = parseURIReference x
-- 								putStrLn $ "==> redirected " ++ (show url) ++ " -> " ++ (show to)
								(text, effurl, headers) <- internalKolRequest to Nothing (cookie, useragent, host, getconn) noredirect
								return (text, effurl, addheaders ++ headers)
							Just to -> do
-- 								putStrLn $ "==> redirected " ++ (show url) ++ " -> " ++ (show to)
								(text, effurl, headers) <- internalKolRequest to Nothing (cookie, useragent, host, getconn) noredirect
								return (text, effurl, addheaders ++ headers)
					_ -> throwIO $ InternalError $ "Error parsing redirect: No location header"
			_ -> do
-- 				putStrLn $ "returning " ++ (show absuri)
				return (body, absuri, hdrs)

internalKolRequest_pipelining :: RefType -> URI -> Maybe [(String, String)] -> Bool -> IO (IO (Data.ByteString.ByteString, URI, [(String, String)]), IO (MVar (Maybe (String, String))))
internalKolRequest_pipelining ref uri params should_invalidate_cache = do
--	putStrLn $ "pipeline: " ++ show uri
	let host = hostUri $ connection $ ref

	curjsonmv <- if should_invalidate_cache
		then do
			doHTTP_DEBUG $ "making new statusmvar for " ++ show uri
			newmv <- newMVar Nothing
			writeIORef (jsonStatusPageMVarRef $ sessionData $ ref) newmv
			return newmv
		else do
			readIORef (jsonStatusPageMVarRef $ sessionData $ ref)
	retrieval_start <- getCurrentTime
	let (absuri, r) = mkreq (useragent $ connection $ ref) (cookie $ connection $ ref) (fromJust $ uri `relativeTo` host) params True
	mv_x <- newEmptyMVar
	writeChan (getconn $ connection $ ref) (absuri, r, mv_x, ref)

	when should_invalidate_cache $ (void $ forkIO $ refreshstatus ref ("dl:" ++ show uri)) `catch` (\e -> do
		putStrLn $ "status retrieval connection error: " ++ (show (e :: ConnectionException)))

	mv_val <- newEmptyMVar
	forkIO $ do
		val <- try $ do
			(absuri, body, hdrs, code, _) <- do
				x <- (readMVar mv_x) `catch` (\e -> do
					putStrLn $ "httpreq read exception for " ++ (uriPath absuri) ++ ": " ++ (show (e :: SomeException))
					throwIO e)
				case x of
					Right r -> return r
					Left e -> throwIO $ HttpRequestException absuri e
			retrieval_end <- getCurrentTime
			prev_retrieval_end <- readIORef (lastRetrieve $ connection $ ref)
			writeIORef (lastRetrieve $ connection $ ref) retrieval_end
			let showurl = case params of
				Nothing -> show uri
				Just p -> show uri ++ " " ++ show p
			log_retrieval ref showurl (max retrieval_start prev_retrieval_end) retrieval_end

			cookie <- case lookup "Set-Cookie" hdrs of
				Just hdrstr -> do
					let cookie = takeWhile (/= ';') hdrstr
					putStrLn $ "set-cookie_pipe: " ++ cookie
					return $ Just cookie
				Nothing -> return (cookie $ connection $ ref)
			case code of
				(3, _, _) -> do
					let Just x = lookup "Location" hdrs
					let addheaders = filter (\(x, _y) -> (x == "Set-Cookie" || x == "Location")) hdrs
					case parseURI x of
						Nothing -> do
							let remapper x
								| (x == ' ') = '+'
								| otherwise = x
							Just to <- case parseURIReference x of
								Just x -> return $ Just x
								_ -> do -- Workaround for broken KoL redirect, e.g. trade counter-offer doesn't urlencode()
									putStrLn $ "parse failure, invalid URL from server: " ++ (show x)
									return $ parseURIReference (map remapper x)
		-- 						putStrLn $ "==> local redirected " ++ (show absuri) ++ " -> " ++ (show to)
							(y, mvy) <- internalKolRequest_pipelining ref to Nothing should_invalidate_cache
							(a, b, c) <- y
							themv <- mvy
							return ((a, b, addheaders ++ c), themv)
						Just to -> do
		-- 					putStrLn $ "==> remote redirected " ++ (show absuri) ++ " -> " ++ (show to)
							-- TODO: make new getconn and use pipelining
							(a, b, c) <- internalKolRequest to Nothing (cookie, (useragent $ connection $ ref), host, Nothing) False
							return ((a, b, c), curjsonmv)
				_ -> do
		-- 			putStrLn $ "returning " ++ (show absuri)
					return ((body, absuri, hdrs), curjsonmv)
		putMVar mv_val val

	let xf = do
		x <- readMVar mv_val
		case x of
			Right (r, _) -> return r
			Left e -> throwIO (e :: SomeException)

	let mvf = do
		x <- readMVar mv_val
		case x of
			Right (_, mv) -> return mv
			Left e -> throwIO (e :: SomeException)

	return (xf, mvf)
