module KolproxyServer where

import Prelude hiding (read, catch)

import HardcodedGameStuff
import PlatformLowlevel
import Logging

import KoL.Util
import KoL.UtilTypes
import KoL.HttpLowlevel
import KoL.Api
import System.Random

import Data.Maybe
import Data.IORef
import Data.Time

import Control.Monad
import Control.Concurrent
import Control.Exception
import Control.Applicative

import Network.CGI (formDecode)
import Network.URI
import Network.HTTP

import System.IO

import qualified Network.Socket
import qualified Data.Map
import qualified Data.ByteString.Char8

import Text.Regex.TDFA

doSERVER_DEBUG _ = return ()
-- doSERVER_DEBUG x = putStrLn $ "SERVER DEBUG: " ++ x

handle_connection sessionmastermv mvsequence mvwhenever h logchan dropping_logchan logcontextmv sh dblogstuff statestuff globalref = do
	recvdata <- kolproxy_receiveHTTP h
	doSERVER_DEBUG "> got request"
	case recvdata of
		Left err -> putStrLn $ "recverror " ++ (show err)
		Right req -> do
--			putStrLn $ "req: " ++ show req
			case uriPath $ rqURI $ req of
				"/kolproxy-test" -> do
					resp <- makeResponseWithNoExtraHeaders (Data.ByteString.Char8.pack "Kolproxy is alive.") "/kolproxy-test" [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")]
					send_http_response h resp
					end_http h
				"/kolproxy-shutdown" -> do
					let maybeparams = decodeUrlParams $ rqURI $ req
					let verified = case maybeparams of
						Just params -> case lookup "secretkey" params of
							Just x -> x == (shutdown_secret $ globalref)
							_ -> False
						_ -> False
					if verified
						then do
							writeIORef (shutdown_ref $ globalref) True
							resp <- makeResponseWithNoExtraHeaders (Data.ByteString.Char8.pack "<html><body><span style=\"color: green\">Closing.</span></body></html>") "/kolproxy-shutdown" [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]
							send_http_response h resp
							end_http h
						else do
							resp <- makeResponseWithNoExtraHeaders (Data.ByteString.Char8.pack "<html><body><span style=\"color: darkorange\">Denied.</span></body></html>") "/kolproxy-shutdown" [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]
							send_http_response h resp
							end_http h
				_ -> void $ forkIO $ do
					doSERVER_DEBUG $ "got request: " ++ (show $ rqURI req) ++ " [" ++ show sh ++ "]"
					let cookie = findHeader HdrCookie req
					let Just uri = parseURIReference $ modded (show $ rqURI req)
						where modded x = if take 2 x == "//" then modded (tail x) else x
					let params = case rqBody req of
						"" -> Nothing
						_ -> Just $ formDecode $ rqBody req

					let useragent_real = kolproxy_version_string ++ " (" ++ platform_name ++ ")" ++ (maybe "" (" " ++) $ findHeader HdrUserAgent req)
					let useragent = if (useragent_real =~ "Safari") && (not (useragent_real =~ "Chrome"))
						then useragent_real ++ " NginxSafariBugWorkaround/1.0 (Faking Chrome/version)"
						else useragent_real

					kolproxy_host <- fromMaybe "http://www.kingdomofloathing.com/" <$> getEnvironmentSetting "KOLPROXY_SERVER"
					use_proxy <- getEnvironmentSetting "KOLPROXY_USE_PROXY_SERVER" -- TODO: merge with doHTTPreq code
					let kolproxy_direct_connection = case use_proxy of
						Nothing -> kolproxy_host
						Just p -> "proxy://" ++ p ++ "/"

					sc <- do
						let m_id = (get_sessid cookie, kolproxy_direct_connection)
						sc <- modifyMVar sessionmastermv (\m -> do
							doSERVER_DEBUG $ "m: " ++ (show $ Data.Map.keys m)
							case Data.Map.lookup m_id m of
								Just x -> do
									doSERVER_DEBUG $ "returning (m, x) " ++ show m_id
									return (m, x)
								Nothing -> do
									doSERVER_DEBUG $ "+++ making new (m, x) +++ " ++ show m_id
									cs <- mkconnthing kolproxy_direct_connection
									cw <- mkconnthing kolproxy_direct_connection
									jspmv <- newMVar Nothing
									jspmvref <- newIORef jspmv
									lrjtref <- newIORef Nothing
									tnow <- getCurrentTime
									lrs <- newIORef tnow
									lrw <- newIORef tnow
									idata <- newIORef Nothing
									statedata <- newIORef Nothing
									let dologaction ref action = do
										doSERVER_DEBUG "dologaction"
										Just jsontext <- readIORef (latestRawJsonText $ sessionData_ $ otherstuff_ $ ref)
										ai <- rawGetApiInfo ref jsontext
										dblogstuff ((charName ai) ++ "-" ++ (show $ ascension ai) ++ ".ascension-log.sqlite3") action
									let dostateaction ref action = do
										doSERVER_DEBUG "dostateaction"
										Just jsontext <- readIORef (latestRawJsonText $ sessionData_ $ otherstuff_ $ ref)
										ai <- rawGetApiInfo ref jsontext
										statestuff ("character-" ++ (charName ai) ++ ".state.sqlite3") action
									luainstancesref <- newIORef $ Data.Map.empty
									laststoredstateref <- newIORef Nothing
									let sessdata = SessionDataType {
										jsonStatusPageMVarRef = jspmvref,
										latestRawJsonText = lrjtref,
										itemData = idata,
										doDbLogAction = dologaction,
										doStateAction = dostateaction,
										stateData = statedata,
										luaInstances = luainstancesref,
										lastStoredState = laststoredstateref
									}
									let sessionconn = ServerSessionType {
										sequenceConnection = cs,
										wheneverConnection = cw,
										sequenceLastRetrieve = lrs,
										wheneverLastRetrieve = lrs,
										sessConnData = sessdata
									}
									return (Data.Map.insert m_id sessionconn m, sessionconn))
						return sc

					let useWhenever = case uriPath uri of
						"/favicon.ico" -> True
						"/newchatmessages.php" -> True
						"/submitnewchat.php" -> True -- TODO: lua-chatcommands
						_ -> False

					let baseref = RefType {
						logstuff_ = LogRefStuff { logchan_ = if useWhenever then dropping_logchan else logchan, logcontext_ = logcontextmv, solid_logchan_ = logchan },
						processingstuff_ = undefined,
						otherstuff_ = OtherRefStuff {
							connection_ = ConnectionType {
								cookie = cookie,
								useragent = useragent,
								KoL.UtilTypes.hostUri = fromJust $ parseURI kolproxy_host,
								lastRetrieve = if useWhenever then wheneverLastRetrieve sc else sequenceLastRetrieve sc,
								connLogSymbol = if useWhenever then "w" else "s",
								getconn = if useWhenever then wheneverConnection sc else sequenceConnection sc
							},
							sessionData_ = sessConnData sc
						},
						stateValid_ = False,
						globalstuff_ = globalref
					}

					mymv <- newEmptyMVar
					writeChan (if useWhenever then mvwhenever else mvsequence) (uri, params, baseref, mymv)
					eitherresp <- takeMVar mymv

					let finalresp = case eitherresp of
						Right resp -> resp
						Left err -> mkResponse (5,0,0) [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")] $ Data.ByteString.Char8.pack $ "Error:\n\n" ++ show err

					send_http_response h finalresp
					end_http h

runProxyServer r rwhenever portnum = do
	logchan <- newChan
	forkIO $ forever $ ((do
		x <- readChan logchan
		x) `catch` (\e -> do
			putStrLn $ "writelog error: " ++ (show (e :: SomeException))))
	dropping_logchan <- newChan
	forkIO $ forever $ readChan dropping_logchan
	logcontextmv <- newMVar ""
	let logref = LogRefStuff { logchan_ = logchan, logcontext_ = logcontextmv, solid_logchan_ = logchan }
	globalref <- do
		indentref <- newIORef 0
		blockluaref <- newIORef False
		hfiles <- openFile "logs/info/files-downloaded.txt" AppendMode
		hSetBuffering hfiles LineBuffering
		htiming <- openFile "logs/info/timing-log.txt" AppendMode
		hSetBuffering htiming LineBuffering
		hlua <- openFile "logs/info/lua-log.txt" AppendMode
		hSetBuffering hlua LineBuffering
		hhttp <- openFile "logs/info/http-log.txt" AppendMode
		hSetBuffering hhttp LineBuffering
		shutdown_secret <- get_md5 <$> show <$> (randomIO :: IO Integer)
		shutdown_ref <- newIORef False
		return GlobalRefStuff {
			logindents_ = indentref,
			blocking_lua_scripting_ = blockluaref,
			h_files_downloaded_ = hfiles,
			h_timing_log_ = htiming,
			h_lua_log_ = hlua,
			h_http_log_ = hhttp,
			shutdown_secret = shutdown_secret,
			shutdown_ref = shutdown_ref
		}
	-- TODO: get rid of faking here!
	let _fake_other = OtherRefStuff {
		connection_ = undefined,
		sessionData_ = undefined
	}

	let _log_fakeref = RefType { logstuff_ = logref, processingstuff_ = undefined, otherstuff_ = _fake_other, stateValid_ = undefined, globalstuff_ = globalref }
	log_debug_msg _log_fakeref $ "runproxyserver"
	sock <- bracketOnError mksocket (Network.Socket.sClose) (\s -> do
		Network.Socket.setSocketOption s Network.Socket.ReuseAddr 1
		Network.Socket.bindSocket s (Network.Socket.SockAddrInet (fromIntegral portnum) Network.Socket.iNADDR_ANY)
		Network.Socket.listen s Network.Socket.maxListenQueue
		return s)
	mvsequence <- newChan
	forkIO $ forever $ do
		(uri, params, cookie, mvresp) <- readChan mvsequence
-- 		writeIORef iref 0
		resp <- (try $ do
			log_file_retrieval _log_fakeref ("proxy:" ++ show uri) params
			log_time_uri _log_fakeref "request" uri
			resp <- log_time_interval _log_fakeref ("creating response for: " ++ (show uri)) $ r uri params cookie
			return resp) :: IO (Either SomeException (Response Data.ByteString.Char8.ByteString))
		putMVar mvresp resp
	mvwhenever <- newChan
	forkIO $ forever $ do
		(uri, params, cookie, mvresp) <- readChan mvwhenever
		forkIO $ do -- Serve asynced, possible private-message loss
-- 		do -- Serve synced, possible hang on timeouts
			resp <- try $ rwhenever uri params cookie
			putMVar mvresp resp
	sessionmastermv <- newMVar Data.Map.empty

	dblogmapmv <- newMVar Data.Map.empty
	let dblogstuff filename action = do
		chan <- modifyMVar dblogmapmv (\m -> do
			case Data.Map.lookup filename m of
				Just x -> do
					return (m, x)
				Nothing -> do
					putStrLn $ "opening log db: " ++ filename
					opendb <- create_db "sqlite3 log" filename
					do_db_query opendb "CREATE TABLE IF NOT EXISTS pageloads(idx INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, time TEXT NOT NULL, statusbefore TEXT NOT NULL, statusafter TEXT, statebefore TEXT, stateafter TEXT, sessionid TEXT NOT NULL, requestedurl TEXT NOT NULL, parameters TEXT, retrievedurl TEXT, pagetext TEXT);" []
					x <- newChan
					forkIO $ forever $ do
						action <- readChan x
						action opendb
					return (Data.Map.insert filename x m, x))
		writeChan chan action

	statemapmv <- newMVar Data.Map.empty
	let statestuff filename action = do
		chan <- modifyMVar statemapmv (\m -> do
			case Data.Map.lookup filename m of
				Just x -> do
					return (m, x)
				Nothing -> do
					putStrLn $ "opening state db: " ++ filename
					statedb <- create_db "state" filename
					x <- newChan
					forkIO $ forever $ do
						action <- readChan x
						action statedb
					return (Data.Map.insert filename x m, x))
		writeChan chan action

	launched_ref <- newIORef False
	forkIO $ do
		update_data_files
		writeIORef launched_ref True
		envlaunch <- getEnvironmentSetting "KOLPROXY_LAUNCH_BROWSER"
		case envlaunch of
			Just "0" -> return ()
			_ -> platform_launch portnum
	let do_loop = do
		should_stop <- readIORef $ shutdown_ref $ globalref
		if should_stop
			then return ()
			else do
				doSERVER_DEBUG "listening on socket"
				(sh, _) <- Network.Socket.accept sock
				h <- socketConnection "???" sh
				doSERVER_DEBUG $ "doing socket:" ++ (show sh)
				launched <- readIORef launched_ref
				if launched
					then ((handle_connection sessionmastermv mvsequence mvwhenever h logchan dropping_logchan logcontextmv sh dblogstuff statestuff globalref) `catch` (\e -> do
						putStrLn $ "proxyError IO: " ++ (show (e :: IOException))) `catch` (\e -> do
							putStrLn $ "proxyError Some: " ++ (show (e :: SomeException))
							throwIO e))
					else do
						kolproxy_receiveHTTP h
						let resp = mkResponse (5,0,3) [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")] $ Data.ByteString.Char8.pack "Error: Kolproxy is still loading data files. Try again in a minute."
						send_http_response h resp
						end_http h
				doSERVER_DEBUG "...done with socket"
				do_loop
	do_loop
	putStrLn "Shutting down."

mkResponse code hdrs text = Response code "" (map (\(x,y) -> mkHeader (HdrCustom x) y) hdrs) text

makeResponse text _effuri headers = return $ mkResponse (2,0,0) (headers ++ [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]) text

makeResponseWithNoExtraHeaders text _effuri headers = return $ mkResponse (2,0,0) headers text

makeRedirectResponse text _effuri headers = return $ mkResponse (3,0,2) (headers ++ [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]) text
